/* -----------------------------------------------------------------------------
 - Project:   Demonstration of basic movements for UM-Dearborn Engineering  
              Experience Day: Animated Robotic Crawlers (ARC) Challenge
 - Copyright: Original software Copyright (C) 2015 SunFounder, 2015/7/21
 -            Modifided by S. Rawashdeh for ARCC, 2017/3/22
 * -----------------------------------------------------------------------------
 - Overview
  - This project was written for the Crawling robot desigened by Sunfounder.
    This version of the robot has 4 legs, and each leg is driven by 3 servos.
    This robot is driven by a Ardunio Nano Board with an expansion Board.
    We recommend that you view the product documentation before using. 
 - Request
  - This project requires some library files, which you can find in the head of
    this file. Make sure you have installed these files.
 

/* Includes ------------------------------------------------------------------*/
#include <Servo.h>		//to define and control servos
#include <FlexiTimer2.h>        //to set a timer to manage all servos
#include <EEPROM.h>		//to save errors of all servos
#include <SPI.h>		//nRF24L01 module need 1/3
#include <nRF24L01.h>	        //nRF24L01 module need 2/3
#include <RF24.h>		//nRF24L01 module need 3/3

/* Calibration/Adjustment values -----------------------------------------------*/
const float adjust_site[3] = { 100, 80, 42 };
const float real_site[4][3] = { { 100, 80, 42 }, { 100, 80, 42 },
                                { 100, 80, 42 }, { 100, 80, 42 } };
/* Servos --------------------------------------------------------------------*/
//define 12 servos for 4 legs
Servo servo[4][3];
//define servos' ports
const int servo_pin[4][3] = { 2, 3, 4, 5, 6, 7, 14, 15, 16, 17, 18, 19 };
/* Wireless communication ----------------------------------------------------*/
//dfine RF24 for nRF24l01
RF24 radio(9, 10);	
//define RF24 transmit pipe
//NOTE: the "LL" at the end of the constant is "LongLong" type
const uint64_t pipe = 0xE8E8F0F0E1LL;
/* Size of the robot ---------------------------------------------------------*/
const float length_a = 55;
const float length_b = 80;
const float length_c = 22.75;
const float length_side = 66;
const float z_absolute = -15;
/* Constants for movement ----------------------------------------------------*/
//const float z_default = -25, z_boot = z_absolute;
//const float x_default = 55;


const float z_default = -65, z_up = 50, z_boot = z_absolute;
//const float z_default = 0, z_up = 15, z_boot = z_absolute;
const float x_default = 60; 
const float y_start = 0, y_step = 50;
const float y_default = x_default;

 
/* variables for movement ----------------------------------------------------*/
volatile float site_now[4][3];      //real-time coordinates of the end of each leg 
volatile float site_expect[4][3];   //expected coordinates of the end of each leg
float temp_speed[4][3];	            //each axis' speed, needs to be recalculated before each movement
float move_speed = 1.4;	            //movement speed
float speed_multiple = 1;           //movement speed multiple
//functions' parameter
const float KEEP = 255;

// Variables for Crawling
const float leg_move_speed = 8;
const float body_move_speed = 3;
/* ---------------------------------------------------------------------------*/

/*
 - setup function
 * ---------------------------------------------------------------------------*/
void setup()
{

	//start serial for debug
	Serial.begin(115200);
	Serial.println("Robot starts initialization");
	//start listen radio
	radio.begin();
	radio.openReadingPipe(1,pipe);
	radio.setRetries(0, 15);
	radio.setPALevel(RF24_PA_HIGH);
	radio.startListening();
	Serial.println("Radio listening started");

	//Move robot to sitting pose
	set_site(0, x_default, y_default, z_boot);
	set_site(1, x_default, y_default, z_boot);
	set_site(2, x_default, y_default, z_boot);
	set_site(3, x_default, y_default, z_boot);
	
	for (int i = 0; i < 4; i++) //for all 4 legs
	{
		for (int j = 0; j < 3; j++) //for all 3 motors
		{
			site_now[i][j] = site_expect[i][j];
		}
	}
 
	//start servo service
	FlexiTimer2::set(20, servo_service);
	FlexiTimer2::start();
	Serial.println("Servo service started");
	//initialize servos
	for (int i = 0; i < 4; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			servo[i][j].attach(servo_pin[i][j]);
			delay(100);
		}
	}
	Serial.println("Servos initialized");
	Serial.println("Robot initialization Complete");
}

/*
 - loop function
 * ---------------------------------------------------------------------------*/
void loop()
{

//put your code here ---------------------------------------------------------
	Serial.println("Welcome to the ARC Challenge 2017!");

#define move_length 40
#define move_up 15

  move_body_absolute(0, 0, move_up);  // starting position, centerd but raised (positive value on Z-axis
  wait_all_reach();
  
// *****************************************************************************************  
// This code segment is a "dance". With legs planted, the body moves left, back, right, then forward.
// *****************************************************************************************
	move_body_absolute(-move_length, 0, move_up);  //moves left: negative on the X-xis
	move_body_absolute(0, -move_length, move_up);  //moves back: negative on Y-axis
	move_body_absolute(move_length, 0, move_up);   // moves right: positive on X-axis
	move_body_absolute(0, move_length, move_up);  // moves fowrard: positive on Y-axis
  move_body_absolute(0, 0, move_up);            // moves center: 0 on bot X and Y axes

  delay(1000);

  // sitting to standing sequence
  move_body_absolute(0, 0, 0);          
  wait_all_reach();   
  move_body_absolute(0, 0, move_up); 
  wait_all_reach();   
  move_body_absolute(0, 0, 0);       
  wait_all_reach();   
  move_body_absolute(0, 0, move_up);    
  wait_all_reach();
  move_body_absolute(0, 0, 0);          
  wait_all_reach();   


  delay(2000);


// *****************************************************************************************  
// Demonstration of rotation (nodding)
// *****************************************************************************************  

#define rotate_degree 20

	move_body_absolute(0, 0, move_up);
	delay(200);

	rotate_body_absolute_x(-rotate_degree, move_up);  // Nose down (negative rotation on X)
	rotate_body_absolute_x(rotate_degree, move_up);   // Nose up (positive rotation on X)
	move_body_absolute(0, 0, move_up);
	rotate_body_absolute_y(-rotate_degree, move_up); // Lean left (negative vlalue on Y)
	rotate_body_absolute_y(rotate_degree, move_up);  // Lean right (positive value on y)
	move_body_absolute(0, 0, move_up);

	move_body_absolute(0, 0, 0);
	wait_all_reach();
	delay(2000);


// *****************************************************************************************  
// Demonstration of crawling
// *****************************************************************************************  

      while(1)
      {

////////////////////////////////////////////////////      
// PHASE 1 of gait: forward-left and rear-right legs

      //Move LEG 2 (forward left)
      move_speed = leg_move_speed;
      set_site(2, x_default, 0, z_up);                // raise it
      wait_all_reach();
      set_site(2, x_default, 2 * y_step, z_up);       // move it fowrard
      wait_all_reach();
      set_site(2, x_default, 2 * y_step, z_default);  // put it down
      wait_all_reach();

      //Move body forward - flexing LEG 2, and extending LEG 1
      move_speed = body_move_speed;
      set_site(0, x_default, 0, z_default);
      set_site(1, x_default, 2 * y_step, z_default);  // note LEG 1 is extended to 2*y_step, we will bring it in next
      set_site(2, x_default, y_step, z_default);
      set_site(3, x_default, y_step, z_default);
      wait_all_reach();

      // Move LEG 1  (rear right)
      move_speed = leg_move_speed;
      set_site(1, x_default, 2 * y_step, z_up);       // raise it (it's extended from the previous step)
      wait_all_reach();
      set_site(1, x_default, 0, z_up);                // bring it in
      wait_all_reach();
      set_site(1, x_default, 0, z_default);           // put it down
      wait_all_reach();

////////////////////////////////////////////////////
// PHASE 2 of gait, forward-right and rear-left legs

      //Move LEG 0 (forward right)
      set_site(0, x_default, 0, z_up);                // raise it
      wait_all_reach();
      set_site(0, x_default, 2 * y_step, z_up);       // move it forward
      wait_all_reach();
      set_site(0, x_default, 2 * y_step, z_default);  // put it down  
      wait_all_reach();

      //Move body forward - flexing LEG 0, and extending LEG 3
      move_speed = body_move_speed;
      set_site(0, x_default, y_step, z_default);
      set_site(1, x_default, y_step, z_default);
      set_site(2, x_default, 0, z_default);
      set_site(3, x_default, 2 * y_step, z_default);
      wait_all_reach();

      // Move LEG 3 (rear left)
      move_speed = leg_move_speed;  
      set_site(3, x_default, 2 * y_step, z_up);       //  raise it (it's extended from previous step)
      wait_all_reach();
      set_site(3, x_default, 0, z_up);                //  bring it in
      wait_all_reach();
      set_site(3, x_default, 0, z_default);           //  put it down
      wait_all_reach();

      }
 

} //end of your code -----------------------------------------------------------




// *****************************************************************************************  
// ARCC COMPETTITORS:
// EDIT AT YOUR OWN RISK: You should not need to edit the code below for the competition. 
// You're encouraged to study the software at another time. It consists of low the low level 
// processor and mathematical conversions that achieve the functionality demonstrated above
// *****************************************************************************************  

/*
 - move body function
 * ---------------------------------------------------------------------------*/
void move_body_relative(float dx, float dy, float dz)
{
  set_site(0, site_now[0][0] - dx, site_now[0][1] - dy, site_now[0][2] - dz);
  set_site(1, site_now[1][0] - dx, site_now[1][1] + dy, site_now[1][2] - dz);
  set_site(2, site_now[2][0] + dx, site_now[2][1] - dy, site_now[2][2] - dz);
  set_site(3, site_now[3][0] + dx, site_now[3][1] + dy, site_now[3][2] - dz);
  wait_all_reach();
}

/*
 - rotate body by x axis function
 * ---------------------------------------------------------------------------*/
void rotate_body_absolute_x(float degree_x, float z)
{
	degree_x = degree_x * PI / 180;
	float dz = (length_side / 2 + y_default)*sin(degree_x);
	float dy = (length_side / 2 + y_default)*(1 - cos(degree_x));

	set_site(0, x_default, y_default - dy, z_default - z - dz);
	set_site(1, x_default, y_default - dy, z_default - z + dz);
	set_site(2, x_default, y_default - dy, z_default - z - dz);
	set_site(3, x_default, y_default - dy, z_default - z + dz);
	wait_all_reach();
}

/*
 - rotate body by y axis function
 * ---------------------------------------------------------------------------*/
void rotate_body_absolute_y(float degree_y, float z)
{
	degree_y = degree_y * PI / 180;
	float dz = (length_side / 2 + x_default)*sin(degree_y);
	float dx = (length_side / 2 + x_default)*(1 - cos(degree_y));

	set_site(0, x_default - dx, y_default, z_default - z + dz);
	set_site(1, x_default - dx, y_default, z_default - z + dz);
	set_site(2, x_default - dx, y_default, z_default - z - dz);
	set_site(3, x_default - dx, y_default, z_default - z - dz);
	wait_all_reach();
}


/*
 - move body function
 * ---------------------------------------------------------------------------*/
void move_body_absolute(float x, float y, float z)
{
	set_site(0, x_default - x, y_default - y, z_default - z);
	set_site(1, x_default - x, y_default + y, z_default - z);
	set_site(2, x_default + x, y_default - y, z_default - z);
	set_site(3, x_default + x, y_default + y, z_default - z);
	wait_all_reach();
}

/*
 - set one of end points' expect site 
 - this founction will set temp_speed[4][3] at same time 
 - non - blocking function
 * ---------------------------------------------------------------------------*/
void set_site(int leg, float x, float y, float z)
{
	float length_x = 0, length_y = 0, length_z = 0;

	if (x != KEEP)
		length_x = x - site_now[leg][0];
	if (y != KEEP)
		length_y = y - site_now[leg][1];
	if (z != KEEP)
		length_z = z - site_now[leg][2];

	float length = sqrt(pow(length_x, 2) + pow(length_y, 2) + pow(length_z, 2));

	temp_speed[leg][0] = length_x / length * move_speed*speed_multiple;
	temp_speed[leg][1] = length_y / length * move_speed*speed_multiple;
	temp_speed[leg][2] = length_z / length * move_speed*speed_multiple;

	if (x != KEEP)
		site_expect[leg][0] = x;
	if (y != KEEP)
		site_expect[leg][1] = y;
	if (z != KEEP)
		site_expect[leg][2] = z;
}

/*
 - wait one of end points move to expect site 
 - blocking function 
 * ---------------------------------------------------------------------------*/
void wait_reach(int leg)
{
	while (1)
		if (site_now[leg][0] == site_expect[leg][0])
			if (site_now[leg][1] == site_expect[leg][1])
				if (site_now[leg][2] == site_expect[leg][2])
					break;
}

/*
 - wait one of end points move to one site 
 - blocking function 
 * ---------------------------------------------------------------------------*/
void wait_reach(int leg, float x, float y, float z)
{
	while (1)
		if (site_now[leg][0] == x)
			if (site_now[leg][1] == y)
				if (site_now[leg][2] == z)
					break;
}

/*
 - wait all of end points move to expect site 
 - blocking function 
 * ---------------------------------------------------------------------------*/
void wait_all_reach(void)
{
	for (int i = 0; i < 4; i++)
		wait_reach(i);
}

/*
 - trans site from polar to cartesian 
 - mathematical model 1/2
 * ---------------------------------------------------------------------------*/
void polar_to_cartesian(volatile float alpha, volatile float beta, volatile float gamma, volatile float &x, volatile float &y, volatile float &z)
{
	//trans degree 180->PI
	alpha = alpha / 180 * PI;
	beta = beta / 180 * PI;
	gamma = gamma / 180 * PI;
	//calculate w-z position
	float v, w;
	v = length_a*cos(alpha) - length_b*cos(alpha + beta);
	z = length_a*sin(alpha) - length_b*sin(alpha + beta);
	w = v + length_c;
	//calculate x-y-z position
	x = w*cos(gamma);
	y = w*sin(gamma);
}

/*
 - trans site from cartesian to polar 
 - mathematical model 2/2
 * ---------------------------------------------------------------------------*/
void cartesian_to_polar(volatile float &alpha, volatile float &beta, volatile float &gamma, volatile float x, volatile float y, volatile float z)
{
	//calculate w-z degree
	float v, w;
	w = (x >= 0 ? 1 : -1)*(sqrt(pow(x, 2) + pow(y, 2)));
	v = w - length_c;
	alpha = atan2(z, v) + acos((pow(length_a, 2) - pow(length_b, 2) + pow(v, 2) + pow(z, 2)) / 2 / length_a / sqrt(pow(v, 2) + pow(z, 2)));
	beta = acos((pow(length_a, 2) + pow(length_b, 2) - pow(v, 2) - pow(z, 2)) / 2 / length_a / length_b);
	//calculate x-y-z degree
	gamma = (w >= 0) ? atan2(y, x) : atan2(-y, -x);
	//trans degree PI->180
	alpha = alpha / PI * 180;
	beta = beta / PI * 180;
	gamma = gamma / PI * 180;
}

/*
 - trans site from polar to microservos
 - mathematical model map to fact
 - the errors saved in eeprom will be add
 * ---------------------------------------------------------------------------*/
void polar_to_servo(int leg, float alpha, float beta, float gamma)
{
	float alpha_error = EEPROM.read(leg * 6 + 0) - 100 + ((float)EEPROM.read(leg * 6 + 1) - 100) / 100;
	float beta_error  = EEPROM.read(leg * 6 + 2) - 100 + ((float)EEPROM.read(leg * 6 + 3) - 100) / 100;
	float gamma_error = EEPROM.read(leg * 6 + 4) - 100 + ((float)EEPROM.read(leg * 6 + 5) - 100) / 100;

	alpha += alpha_error;
	beta += beta_error;
	gamma += gamma_error;

	if (leg == 0)
	{
		alpha = 90 - alpha;
		beta = beta;
		gamma = 90 - gamma;
	}
	else if (leg == 1)
	{
		alpha += 90;
		beta = 180 - beta;
		gamma += 90;
	}
	else if (leg == 2)
	{
		alpha += 90;
		beta = 180 - beta;
		gamma += 90;
	}
	else if (leg == 3)
	{
		alpha = 90 - alpha;
		beta = beta;
		gamma = 90 - gamma;
	}

	servo[leg][0].write(alpha);
	servo[leg][1].write(beta);
	servo[leg][2].write(gamma);
}

/*
 - microservos service /timer interrupt function/50Hz
 - when set site expected,this function move the end point to it in a straight line
 - temp_speed[4][3] should be set before set expect site,it make sure the end point
   move in a straight line,and decide move speed.
 * ---------------------------------------------------------------------------*/
void servo_service(void)
{
  sei();
  static float alpha, beta, gamma;

  for (int i = 0; i < 4; i++)
  {
    for (int j = 0; j < 3; j++)
    {
      if (abs(site_now[i][j] - site_expect[i][j]) >= abs(temp_speed[i][j]))
        site_now[i][j] += temp_speed[i][j];
      else
        site_now[i][j] = site_expect[i][j];
    }

    cartesian_to_polar(alpha, beta, gamma, site_now[i][0], site_now[i][1], site_now[i][2]);
    polar_to_servo(i, alpha, beta, gamma);
  }
}
